@@ -41,6 +41,7 @@
 #include "firebuild/epoll.h"
 #include "firebuild/exe_matcher.h"
 #include "firebuild/file_name.h"
+#include "firebuild/file_usage_update.h"
 #include "firebuild/execed_process.h"
 #include "firebuild/execed_process_cacher.h"
 #include "firebuild/hash_cache.h"
@@ -664,7 +665,9 @@ static void process_posix_spawn_file_actions(const P * const ic_msg, Process* co
             int flags = action_open->get_flags();
             mode_t mode = action_open->get_mode();
             proc->handle_force_close(fd);
-            proc->handle_open(AT_FDCWD, pathname, pathname_len, flags, mode, fd, 0, -1, 0,
+            proc->handle_open(AT_FDCWD, pathname, pathname_len, flags, mode,
+                              false /* is_openat2 */, 0 /* resolve_flags */,
+                              fd, 0, -1, 0,
                               is_write(flags), false);
             break;
           }
@@ -768,8 +771,25 @@ static const FileName* resolve_command_from_msg(const T* ic_msg, const char* com
       std::vector<std::string> env = ic_msg->get_env_as_vector();
       size_t path_len = 0;
       const char* path = Env::get_var(env, "PATH", &path_len);
-      return hash_cache->resolve_command(command, command_len, path, path_len,
-                                         proc->wd());
+      std::vector<const FileName*> paths_checked, paths_checked_is_dir;
+      const FileName* result = hash_cache->resolve_command(command, command_len, path, path_len,
+                                                           proc->wd(), &paths_checked,
+                                                           &paths_checked_is_dir);
+      /* Register paths that were searched but did not contain the executable.
+       * These should be registered as NOTEXIST so that if a new executable appears
+       * earlier in PATH, the cache entry will be invalidated. */
+      ExecedProcess* exec_point = proc->exec_point();
+      if (exec_point) {
+        for (const FileName* checked_path : paths_checked) {
+          FileUsageUpdate update(checked_path, NOTEXIST);
+          exec_point->register_file_usage_update(checked_path, update);
+        }
+        for (const FileName* checked_path : paths_checked_is_dir) {
+          FileUsageUpdate update(checked_path, ISDIR);
+          exec_point->register_file_usage_update(checked_path, update);
+        }
+      }
+      return result;
     }
   } else {
     return proc->get_absolute(AT_FDCWD, command, command_len);
@@ -1273,32 +1293,56 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       auto ic_msg = reinterpret_cast<const FBBCOMM_Serialized_futime *>(fbbcomm_buf);
       const int fd = ic_msg->get_fd();
       const FileFD* ffd = proc->get_fd(fd);
+      const FileName* touch_source_file = nullptr;
       if (!ic_msg->has_error_no() && ffd && is_write(ffd->flags()) && ic_msg->get_all_utime_now()) {
         /* The fd has been opened for writing and the access and modification times should be set to
          * current time which happens automatically when the process is shortcut. This is safe. */
-      } else {
-        ExecedProcess* next_exec_level;
-        if (quirks & FB_QUIRK_LTO_WRAPPER && proc->exec_point()->args().size() > 0
-            && (proc->exec_point()->args()[0] == "touch"
-                || (proc->exec_point()->executable()->without_dirs() == "coreutils"
-                    && proc->exec_point()->args().size() > 1
-                    && proc->exec_point()->args()[1] == "--coreutils-prog-shebang=touch"))
-            && (next_exec_level = proc->parent_exec_point())  // sh
-            && (next_exec_level = next_exec_level->parent_exec_point())  // make
-            && (next_exec_level = next_exec_level->parent_exec_point())  // lto-wrapper
-            && next_exec_level->executable()->without_dirs() == "lto-wrapper" ) {
-          FB_DEBUG(FB_DEBUG_PROC, "Allow shortcutting lto-wrapper's touch descendant "
-                   "(lto-wrapper quirk)");
+      } else if (!ic_msg->has_error_no() && !ic_msg->get_all_utime_now()
+                 && ic_msg->has_mtim_sec() && ic_msg->has_mtim_nsec()
+                 && proc->is_touch_r_operation(ic_msg->get_mtim_sec(), ic_msg->get_mtim_nsec(),
+                                               &touch_source_file)) {
+        /* This is a touch -r operation copying timestamps from a file we stat'ed.
+         * Register the file usage with the timestamp source for cache replay. */
+        FB_DEBUG(FB_DEBUG_PROC, "Detected touch -r operation (futime), allowing shortcutting");
+        if (ffd && ffd->filename()) {
+          FileUsageUpdate update(ffd->filename(), DONTKNOW, true);
+          update.set_timestamp_source(touch_source_file);
+          proc->exec_point()->register_file_usage_update(ffd->filename(), update);
         } else {
           proc->exec_point()->disable_shortcutting_bubble_up(
-              "Changing file timestamps is not supported");
+              "Can't file file changing file timestamps of");
         }
+      } else {
+        proc->exec_point()->disable_shortcutting_bubble_up(
+            "Changing file timestamps is not supported");
       }
       break;
     }
     case FBBCOMM_TAG_utime: {
-      proc->exec_point()->disable_shortcutting_bubble_up(
-          "Changing file timestamps is not supported");
+      auto ic_msg = reinterpret_cast<const FBBCOMM_Serialized_utime *>(fbbcomm_buf);
+      const FileName* touch_source_file = nullptr;
+      if (!ic_msg->has_error_no() && !ic_msg->get_all_utime_now()
+          && ic_msg->has_mtim_sec() && ic_msg->has_mtim_nsec()
+          && proc->is_touch_r_operation(ic_msg->get_mtim_sec(), ic_msg->get_mtim_nsec(),
+                                        &touch_source_file)) {
+        /* This is a touch -r operation copying timestamps from a file we stat'ed.
+         * Register the file usage with the timestamp source for cache replay. */
+        FB_DEBUG(FB_DEBUG_PROC, "Detected touch -r operation (utime), allowing shortcutting");
+        if (ic_msg->has_pathname()) {
+          const FileName* target_file =
+              FileName::GetCanonicalized(ic_msg->get_pathname(), ic_msg->get_pathname_len(),
+                                         proc->wd());
+          FileUsageUpdate update(target_file, DONTKNOW, true);
+          update.set_timestamp_source(touch_source_file);
+          proc->exec_point()->register_file_usage_update(target_file, update);
+        } else {
+          proc->exec_point()->disable_shortcutting_bubble_up(
+              "Changing timestamp with utime() without filename");
+        }
+      } else {
+        proc->exec_point()->disable_shortcutting_bubble_up(
+            "Changing file timestamps is not supported");
+      }
       break;
     }
     case FBBCOMM_TAG_clock_gettime: {
