@@ -46,7 +46,9 @@ class ProcessFBBAdaptor {
     int error = msg->get_error_no_with_fallback(0);
     int ret = msg->get_ret_with_fallback(-1);
     return proc->handle_open(dirfd, msg->get_pathname(), msg->get_pathname_len(), msg->get_flags(),
-                             msg->get_mode_with_fallback(0), ret, error, fd_conn, ack_num,
+                             msg->get_mode_with_fallback(0),
+                             msg->has_resolve_flags(), msg->get_resolve_flags_with_fallback(0),
+                             ret, error, fd_conn, ack_num,
                              msg->get_pre_open_sent(), msg->get_tmp_file_with_fallback(false));
   }
 
@@ -120,10 +122,14 @@ class ProcessFBBAdaptor {
     const int fd = msg->get_fd_with_fallback(AT_FDCWD);
     const mode_t st_mode = msg->get_st_mode_with_fallback(0);
     const off_t st_size = msg->get_st_size_with_fallback(0);
+    const int64_t st_mtim_sec = msg->get_st_mtim_sec_with_fallback(0);
+    const int64_t st_mtim_nsec = msg->get_st_mtim_nsec_with_fallback(0);
     const int flags = msg->get_flags_with_fallback(0);
     const int error = msg->get_error_no_with_fallback(0);
     return proc->handle_fstatat(fd, msg->get_pathname(), msg->get_pathname_len(),
-                                flags, st_mode, st_size, error);
+                                flags, st_mode, st_size,
+                                st_mtim_sec, st_mtim_nsec,
+                                error);
   }
 
   static int handle(Process *proc, const FBBCOMM_Serialized_faccessat *msg) {
