@@ -23,8 +23,7 @@
 
 #include "common/platform.h"
 #include "firebuild/debug.h"
-
-extern bool generate_report;
+#include "firebuild/options.h"
 
 namespace firebuild {
 
@@ -85,7 +84,7 @@ void ProcessTree::insert_root(pid_t root_pid, int stdin_fd, int stdout_fd, int s
   root_->set_state(firebuild::FB_PROC_TERMINATED);
   // TODO(rbalint) support other inherited fds
   /* Create the FileFD representing stdin of the top process. Pipe will be NULL, that's fine. */
-  root_->add_filefd(stdin_fd, std::make_shared<FileFD>(stdin_fd, O_RDONLY, nullptr, nullptr));
+  root_->add_filefd(stdin_fd, std::make_shared<FileFD>(O_RDONLY, nullptr, nullptr));
 
   /* Create the Pipes and FileFDs representing stdout and stderr of the top process. */
   // FIXME Make this more generic, for all the received pipes / terminal outputs.
@@ -98,7 +97,7 @@ void ProcessTree::insert_root(pid_t root_pid, int stdin_fd, int stdout_fd, int s
     if (fd == stderr_fd && stdout_stderr_match) {
       /* stdout and stderr point to the same location (changing one's flags did change the
        * other's). Reuse the Pipe object that we created in the loop's first iteration. */
-      root_->add_filefd(fd, std::make_shared<FileFD>(fd, (*root_->fds())[stdout_fd], false));
+      root_->add_filefd(fd, std::make_shared<FileFD>((*root_->fds())[stdout_fd], false));
     } else {
       /* Create a new Pipe for this file descriptor.
        * The fd keeps blocking/non-blocking behaviour, it seems to be ok with epoll.
@@ -116,7 +115,7 @@ void ProcessTree::insert_root(pid_t root_pid, int stdin_fd, int stdout_fd, int s
       FB_DEBUG(FB_DEBUG_PIPE, "created pipe with fd0: " + d(fd) + ", dup()-ed as: " + d(fd_dup));
       /* Top level inherited fds are special, they should not be closed. */
       inherited_fd_pipes_.insert(pipe);
-      root_->add_filefd(fd, std::make_shared<FileFD>(fd, O_WRONLY, pipe, nullptr));
+      root_->add_filefd(fd, std::make_shared<FileFD>(O_WRONLY, pipe, nullptr));
     }
   }
   insert_process(root_);
@@ -126,7 +125,7 @@ void ProcessTree::GcProcesses() {
   bool have_orphan = root()->has_orphan_descendant();
   while (!proc_gc_queue_.empty()) {
     ExecedProcess* proc = proc_gc_queue_.front();
-    if (!generate_report) {
+    if (!Options::generate_report()) {
       // TODO(rbalint) delete subtrees not affected by orphans
       if (proc->parent() && !have_orphan) {
         proc->parent()->set_exec_child(nullptr);
